#! /usr/bin/python
# -*- coding: utf-8 -*-
# vim:set shiftwidth=4 tabstop=4 noexpandtab:

import os
import sys
import shutil
import codecs
import logging
import giraffe.ast
import giraffe.languages

LEVELS = (  logging.ERROR,
			logging.WARNING,
			logging.INFO,
			logging.DEBUG,
			)

HTML_HEADER = \
"""<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en-US" lang="en-US">
  <head>
    <meta http-equiv="content-type" content="application/xhtml+xml; charset=utf-8"/>

    <title>%(namespace)s-%(version)s %(lang)s API Documentation</title>
    <link rel="stylesheet" href="../giraffe.css" type="text/css" />
  </head>
<body>
<h1>%(namespace)s-%(version)s %(lang)s API Documentation</h1>
"""

HTML_FOOTER = \
"""</body>
</html>
"""

def get_css_path ():
	giraffe_dir = os.path.dirname (giraffe.__file__)
	return os.path.join (giraffe_dir, "data", "giraffe.css")

def get_lang_name (lang):
	return lang.__name__.rpartition(".")[2]

def get_output_dir (lang, output_dir):
	return "%s/%s" % (output_dir, get_lang_name (lang))

def get_output_file (ns, lang, output_dir):
	return get_output_dir (lang, output_dir) + "/%s-%s.html" % (ns.name, ns.version)

class GiraffeApp:
	def __init__(self, args, output_dir, debug_mode = False, logging_level = None):
		self.output_dir = output_dir
		self.debug_mode = debug_mode
		if not logging_level:
			# To inherit the logging level from the importer, unless explicitly
			# specified
			self.logging_level = [i for i, lvl in enumerate(LEVELS) if lvl == logging.root.level][0]
		else:
			self.logging_level = logging_level
		self.args = args

		# Set the verbosity
		if self.debug_mode:
			self.logging_level = 3
		logging.basicConfig(level=LEVELS[self.logging_level],
			format='%(asctime)s %(levelname)s %(message)s')

	def run(self):
		# Open and parse input files
		repo = giraffe.ast.Repository()
		for f in self.args:
			print "Reading %s" % f
			repo.add_gir (file(f, "r"))
		repo.link ()

		for lang in giraffe.languages.modules:
			try:
				os.makedirs (get_output_dir (lang, self.output_dir))
			except OSError:
				pass

			for ns in repo.namespaces:
				output_path = get_output_file (ns, lang, self.output_dir)
				output_file = codecs.open (output_path, "w", "utf-8")
				print "Writing %s" % output_path
				header = HTML_HEADER % { "namespace" : ns.name, "version" : ns.version, "lang" : get_lang_name(lang).title() }
				print >> output_file, header
				lang.write_namespace (ns, output_file)
				print >> output_file, HTML_FOOTER

		css_dir = self.output_dir
		css_path = get_css_path ()
		print "Installing %s into %s" % (css_path, css_dir)
		shutil.copy(css_path, css_dir)


