# -*- coding: utf-8 -*-

import giraffe
from giraffe.utils import HTMLWriter, camelcase2lower

ANCHOR=u"⚓"

def write_namespace (ns, stream):
	w = HTMLWriter (stream)
	classes = ns.classes + ns.interfaces
	classes.sort (lambda x, y: cmp(x.c_type, y.c_type))
	for cl in classes:
		write_class (cl, w)

def param_list (params):
	"""Build a list of strings - possibly with markup - for a list of giraffe.ast.Parameters"""
	result = []
	for p in params:
		if p.type.namespaced_name in p.namespace.repository:
			result.append(p.type.html_link(camelcase2lower(p.name)))
		else:
			result.append(camelcase2lower(p.name))
	return result

def type_list (nodes, klass=""):
	"""Build a list of strings - possibly with markup - for a list of giraffe.ast.TypedNodes"""
	result = []
	for n in nodes:
		if n.type.namespaced_name in n.namespace.repository:
			result.append(n.type.html_link(type_label (n.type), klass="type %s" % klass))
		else:
			result.append('<span class="type %s">%s</span>' % (klass, type_label (n.type)))
	return result

def type_label (t):
	if hasattr(t, "pyname"):
		return t.pyname
	else:
		return t.namespaced_name

def write_class (cl, w):
	#
	# Class header
	#
	if isinstance (cl, giraffe.ast.Interface):
		w.start_div (klass="class-interface")
		w.a(ANCHOR, href="#"+cl.anchor, klass="class-anchor")
		w.h2 ("Interface: %s" % cl.namespaced_name, klass="class-interface", id=cl.anchor)				
	elif cl.is_abstract:
		w.start_div (klass="class-abstract")
		w.a(ANCHOR, href="#"+cl.anchor, klass="class-anchor")
		w.h2("Abstract Class: %s" % cl.namespaced_name, klass="class-abstract", id=cl.anchor)
	else:
		w.start_div (klass="class-normal")
		w.a(ANCHOR, href="#"+cl.anchor, klass="class-anchor")
		w.h2("Class: %s" % cl.namespaced_name, klass="class-normal", id=cl.anchor)
	
	#
	# Class relations and interfaces
	#
	w.start_div (klass="relations")
	if isinstance (cl, giraffe.ast.Interface) and cl.implementations:
			w.start_p ()
			w.raw ("Classes implementing %s: %s" % (cl.namespaced_name, ", ".join (map (lambda impl: impl.html_link(), cl.implementations))))
			w.end_p()
	if cl.interfaces:
		w.start_p ()
		w.raw ("Implements: %s" % (", ".join (map (lambda iface: iface.html_link(), cl.interfaces))))
		w.end_p ()
	if cl.subclasses:
		w.start_p ()
		w.raw ("Known subclasses: %s" % (", ".join (map (lambda subclass: subclass.html_link(), cl.subclasses))))
		w.end_p ()
	if cl.parent_class:
		w.start_p ()
		w.raw ("Subclass of: %s" % cl.parent_class.html_link())
		w.end_p ()
	if cl.doc : w.p(cl.doc)
	w.end_div()
	
	#
	# Properties
	#
	if cl.properties :
		w.start_div (klass="properties")
		w.h4("Properties", klass="properties")		
	for prop in cl.properties:
		access = "read-only"
		if prop.construct_only :
			access = "construct-only"
		elif prop.writable:
			access = "read/write"
		w.start_div (klass="property")
		w.start_p()
		w.raw (type_list ([prop], klass="property-type")[0])
		#if prop.type.namespaced_name in cl.namespace.repository:
		#	w.raw(prop.type.html_link(klass="property-type"))
		#else:
		#	w.span(type_label (prop.type), klass="property-type")
		w.span('"%s"' % prop.name, klass="property-name", id=prop.anchor)
		w.span (access, klass="property-access")
		w.end_p()
		if prop.doc:
			w.p (prop.doc, klass="property-doc")
		w.end_div()
	if cl.properties:
		w.end_div()
	
	#
	# Signals
	#
	if cl.signals :
		w.start_div (klass="signals")
		w.h4("Signals")		
	for sig in cl.signals:
		params = param_list (sig.parameters)
		params = ["self", cl.html_link(camelcase2lower(cl.name))] + params
		retvals = type_list ([sig.return_value])
		
		w.start_div (klass="signal")
		w.start_p()
		w.raw ('<span class="signal-return-type">%s</span>' % ", ".join (retvals))
		w.span('"%s"' % sig.name, klass="signal-name", id=sig.anchor)
		w.raw('<span class="signal-parameters">(%s)</span>' % ", ".join (params))
		if sig.doc:
			w.p (sig.doc, klass="signal-doc")
		w.end_p()
		w.end_div()
	if cl.signals:
		w.end_div()
	
	#
	# Constructors
	#
	if cl.constructors :
		w.start_div(klass="methods")
		w.h4 ("Constructors")
	constructors = cl.constructors
	for ctor in constructors:
		if not ctor.introspectable : continue
		
		w.start_div(klass="method constructor")
		w.start_p()
		w.raw ('<span class="method-return-type">&nbsp;</span>')
		w.span(ctor.name, klass="method-name", id=ctor.anchor)
		w.raw('<span class="method-parameters">(%s)</span>' % ", ".join (param_list(ctor.parameters)))
		w.end_p()
		w.end_div()
	if cl.constructors:
		w.end_div()
	
	#
	# Methods
	#	
	if cl.methods :
		w.start_div(klass="methods")
		w.h4 ("Methods")		
	methods = cl.methods
	methods.sort(lambda x,y: cmp(x.c_identifier, y.c_identifier))
	for m in methods:			
		
		if not m.introspectable : continue
		
		params = param_list ([p for p in m.parameters if not p.is_out])
		retvals = type_list([m.return_value])
		extra_retvals = type_list ([p for p in m.parameters if p.is_out])
		retvals += extra_retvals
		
		cssklass = "method"
		if m.virtual : cssklass += " method-virtual"
		if m.static : cssklass += " method-static"
		w.start_div(klass=cssklass)
		w.start_p()
		w.raw ('<span class="method-return-type">%s</span>' % ", ".join (retvals))
		w.span(m.name, klass="method-name", id=m.anchor)
		w.raw('<span class="method-parameters">(%s)</span>' % ", ".join (params))
		if m.static :
			w.span ("(static method)", klass="method-remark")
		w.end_p()
		w.end_div()
	if cl.methods:
		w.end_div()
	
	w.end_div() # End class div


